;;;; -*- mode:Scheme -*- ;;;;


;;;; ASSERTION BASE (WORKING MEMORY)

(define *assertions*)

(define (clear-assertions) (set! *assertions* ()))
 
(define (remember-assertion assertion)
  (if (member assertion *assertions*)
      *assertions*
      (set! *assertions* (cons assertion *assertions*))))

(define (remember-assertions . assertions)
  (for-each remember-assertion assertions))

(define (delete-assertion assertion)
  "
  Purpose:	Deletes assertions that match specified assertion.
  "
  (set! *assertions*
	(append-map (lambda (assert)
		      (if (match assertion assert)
			  (list assert)
			  ()))
		    *assertions*)
	))

(define (count-assertions pattern)
  (length
   (append-map
    (lambda (x)
      (let ((m (match pattern x)))
	(if (eq? m 'fail) () '(#t))))
    *assertions*)))

(define (display-assertions #!optional assertions)
  (if (default-object? assertions) (set! assertions *assertions*))
  (if (not (null? assertions))
      (begin (format #t "~%~a" (first assertions))
	     (display-assertions (cdr assertions)))))

;;;; RULE BASE

(define *rules*)

(define (clear-rules)
  (set! *rules* ()))

(define (remember-rule rule)
  (if (member rule *rules*)
      *rules*
      (set! *rules* (append! *rules* (list rule)))
      ))

(define (remember-rules . rules)
  (for-each remember-rule rules))


;;;; ACCESS FUNCTIONS FOR RULE ELEMENTS

(define (rule-name rule) (first rule))

(define (rule-ifs rule) (extract-from-rule 'if rule))

(define (rule-thens rule) (extract-from-rule 'then rule))

(define (rule-then rule) (first (rule-thens rule)))

(define (rule-adds rule) (extract-from-rule 'add rule))

(define (rule-deletes rule) (extract-from-rule 'delete rule))

(define (rule-and-if rule) (extract-from-rule 'and-if rule))

(define (rule-and-not rule) (extract-from-rule 'and-not rule))

(define (rule-evaluating rule) (extract-from-rule 'evaluating rule))

(define (rule-saying rule) (extract-from-rule 'saying rule))

(define (extract-from-rule marker rule)
  (let ((fragment (member marker rule)))
    (if fragment
	(extract-expressions (cdr fragment))
	())))

(define (extract-expressions rule)
  (cond ((null? rule) ())
	((list? (first rule))
	 (cons (first rule) (extract-expressions (cdr rule))))
	(else ())))

;;; For working on streams of association lists

(define make-a-list-stream list)
(define a-list-stream-first first)
(define a-list-stream-cdr cdr)
(define a-list-stream-cons cons)
(define a-list-stream-append append)
(define a-list-stream-member member)
(define a-list-stream-map map)
(define (a-list-stream-concatenate a-list-streams) (apply append a-list-streams))

(define instantiated-rule-name first)
(define instantiated-rule-thens second)
(define instantiated-rule-deletes third)
(define instantiated-rule-evaluatings fourth)
(define instantiated-rule-sayings fifth)
(define instantiated-rule-rule sixth)

;;;; FORWARD CHAINING

(define (chain)
  "
  Purpose:	Initiate forward chaining.
  "
  (let* ((candidates (try-rules))
	 (c (choose-rule candidates)))
    (format #t "~% Triggered rules: ~s" (map first candidates))
    (if c
	;; perform instantiated rule actions
	(let* ((name (instantiated-rule-name c))
	       (thens (instantiated-rule-thens c))
	       (deletes (instantiated-rule-deletes c))
	       (evaluatings (instantiated-rule-evaluatings c))
	       (sayings (instantiated-rule-sayings c)))
	  (for-each (lambda (x)
		      (format #t "~%Rule ~a adds ~a" name x)
		      (remember-assertion x))
		    thens)
	  (for-each (lambda (x)
		      (format #t "~%Rule ~a deletes ~a" name x)
		      (delete-assertion x))
		    deletes)
	  (for-each (lambda (x) (eval x (the-environment))) evaluatings)
	  (for-each (lambda (x) (format #t "~%") (apply format #t x)) sayings)
	  (chain))
	(format #t "~%No rule can assert anything new"))))


(define (choose-rule instantiated-rules)
  (if (pair? instantiated-rules)
      (list-ref instantiated-rules (random (length instantiated-rules)))
      #f))

(define (try-rules)
  (append-map (lambda (rule) (try-rule rule)) *rules*))

(define (try-rule rule)
  "
  Purpose:	Controls forward chaining.
  Remarks:	Handles IF, THEN, ADD, DELETE, AND-IF, EVALUATING,
		and SAYING.  The SAYING marker arranges for printing
		helpful notes for the user.
  " 
  ;; (format #t "~%Trying rule ~a." (rule-name rule))
  (let ((a-list-stream (apply-filters (rule-ifs rule) (cons () (stream)))))
    (append-map
     (lambda (a-list)
       (let ((deletes (instantiate-variables (rule-deletes rule) a-list))
	     (adds (instantiate-variables (rule-adds rule) a-list))
	     (thens (instantiate-variables (rule-thens rule) a-list))
	     (and-if (instantiate-variables (rule-and-if rule) a-list))
	     (and-not (instantiate-variables (rule-and-not rule) a-list))
	     (saying (instantiate-variables (rule-saying rule) a-list))
	     (evaluating
	      (instantiate-variables (rule-evaluating rule) a-list)))
	 (if (and (or (not (for-all? thens
			     (lambda (x) (member x *assertions*))))
		      (not (for-all? adds
			     (lambda (x) (member x *assertions*))))
		      (there-exists? deletes
			(lambda (x) 
			  (or (member x *assertions*)
			      (there-exists? *assertions*
				(lambda (y) (match x y)))))))
		  (for-all? and-not
			    (lambda (x)
			      (for-all? *assertions*
					(lambda (y)
					  (eq? 'fail (match x y))))))
		  (for-all? and-if (lambda (x) (eval x (the-environment)))))
	     (list (list (rule-name rule)
			 (or thens adds)
			 deletes
			 evaluating
			 saying
			 rule))
	     ())))
     a-list-stream)))

(define (apply-filters patterns a-list-stream)
  "
  Purpose:	Tries to match all patterns to all assertions using
		all binding lists.
  "
  (if (null? patterns)
      a-list-stream
      (apply-filters (cdr patterns)
		     (filter-a-list-stream (first patterns)
					   a-list-stream))))

(define (filter-a-list-stream pattern a-list-stream)
  "
  Purpose:	Tries to match one pattern to all assertions using
		all binding lists.
  "
  (a-list-stream-concatenate
   (a-list-stream-map
    (lambda (bindings)
      (match-pattern-to-assertions pattern bindings))
    a-list-stream)))

(define (match-pattern-to-assertions pattern bindings)
  "
  Purpose:	Tries to match one pattern to all assertions using
		one binding list.
  "
  (a-list-stream-concatenate
   (a-list-stream-map
    (lambda (assertion) (try-assertion pattern assertion bindings))
    *assertions*)))

(define (try-assertion pattern assertion bindings)
  "
  Purpose:	Tries to match one pattern to one assertion.
  "
  (let ((result (match pattern assertion bindings)))
    (if (eq? 'fail result)
        (make-a-list-stream)
        (a-list-stream-cons result (make-a-list-stream)))))

(define (instantiate-variables pattern a-list-stream)
  "
  Purpose:	Replaces variables by their bindings.
  "
  (cond ((not (list? pattern)) pattern)
	((null? pattern) ())
	((eq? '? (first pattern))
	 (if (eq? '_ (second pattern))
	     pattern
	     (let ((binding (assoc (second pattern) a-list-stream)))
	       (if binding
		   (second binding)
		   (error "Unbound variable in rule: " pattern)))))
	(else (cons (instantiate-variables (first pattern) a-list-stream)
		    (instantiate-variables (cdr pattern) a-list-stream)))))



