;;; Robot and Tools

;;; Assertion types:
;;; (on <object>)  -- applies to robot
;;; (holds <object>) -- applies to robot
;;; (at <location>) -- applies to robot
;;; (<object> on <object>)
;;; (<object> at <location>)
;;; (goal: on <object>)
;;; (goal: holds <object>)
;;; (goal: at <location>)
;;; (goal: <object> at <location>)

(define (make-simple-robot-rules)
  (clear-rules)
  (remember-rules
   '(walk/do
     IF     (goal: at (? new-loc))
            (on floor)
	    (at (? old-loc))
     SAYING ("WALK to ~a" (? new-loc))
     ADD    (at (? new-loc))
     DELETE (at (? old-loc))
            (goal: at (? new-loc)))
   '(climb-down/do 
     IF     (goal: on floor)
            (on (? support))
     AND-IF (not (equal? '(? support) 'floor)) ;<-- note vars have to be quoted
     SAYING ("CLIMB down to floor")
     ADD    (on floor)
     DELETE (on (? support))
            (goal: on floor))
   '(climb-up/do
     IF     (goal: on (? object))
            (on floor)
            (at (? loc))
	    ((? object) at (? loc))
     SAYING ("CLIMB onto ~a" (? object))
     ADD    (on (? object))
     DELETE (on floor)
            (goal: on (? object)))
   '(climb-up/walk-subgoal
     IF     (goal: on (? object))
            ((? object) at (? new-loc))
            (at (? old-loc))
     AND-IF (not (equal? '(? old-loc) '(? new-loc))) ;<-- quoted vars
     ADD    (goal: at (? new-loc)))
   '(walk/on-floor-subgoal
     IF     (goal: at (? loc))
            (on (? support))
     AND-IF (not (equal? '(? support) 'floor)) ;<-- quoted vars
     THEN   (goal: on floor))))

(define (make-simple-robot-test)
  (clear-assertions)
  (remember-assertions
   '(at (9 5))
   '(on floor)
   '(table at (5 6))
   '(stool at (3 4))
   '(goal: on table)))

(define (simple-robot-test)
  (make-simple-robot-rules)
  (make-simple-robot-test)
  (display-assertions)
  (chain)
  (display-assertions))


(define (make-full-robot-rules)
  (clear-rules)
  (remember-rules
   ;;; walk to a new location by adding yourself to the new location
   '(walk/do
     IF     (goal: at (? new-loc))
            (on floor)
	    (at (? old-loc))
	    (holds nothing)
     SAYING ("WALK to ~a" (? new-loc))
     ADD    (at (? new-loc))
     DELETE (at (? old-loc))
            (goal: at (? new-loc)))
   ;;; when you are holding an object and walking, you have to 
   ;;; move the object as well.
   '(walk-holding/do
     IF     (goal: at (? new-loc))
            (on floor)
	    (at (? old-loc))
	    (holds (? object))
     AND-IF (not (equal? '(? object) 'nothing))
     SAYING ("WALK to ~a with ~a." (? new-loc) (? object))
     ADD    (at (? new-loc))
            ((? object) at (? new-loc))
     DELETE (at (? old-loc))
            ((? object) at (? old-loc))
	    (goal: at (? new-loc)))
   ;;; if you want to walk, move to the floor first
   '(walk/on-floor-subgoal
     IF     (goal: at (? loc))
            (on (? support))
     AND-IF (not (equal? '(? support) 'floor)) ;<-- quoted vars
     THEN   (goal: on floor))
   ;;; climb down to the floor if not on it
   '(climb-down/do 
     IF     (goal: on floor)
            (on (? support))
     AND-IF (not (equal? '(? support) 'floor)) ;<-- note vars have to be quoted
     SAYING ("CLIMB down to floor")
     ADD    (on floor)
     DELETE (on (? support))
            (goal: on floor))
   ;;; climb up to a surface if on the floor and not holding anything
   '(climb-up/do
     IF     (goal: on (? object))
            (on floor)
            (at (? loc))
	    ((? object) at (? loc))
	    (holds nothing)
     SAYING ("CLIMB onto ~a" (? object))
     ADD    (on (? object))
     DELETE (on floor)
            (goal: on (? object)))
   ;;; walk to a location to climb the object at that location
   '(climb-up/walk-subgoal
     IF     (goal: on (? object))
            ((? object) at (? new-loc))
            (at (? old-loc))
     AND-IF (not (equal? '(? old-loc) '(? new-loc))) ;<-- quoted vars
     ADD    (goal: at (? new-loc)))
   ;;; grasp an object on the floor
   '(grasp-on-floor/do
     IF     (goal: holds (? object))
            (on floor)
	    (at (? loc))
	    ((? object) at (? loc))
	    ((? object) on floor)
	    (holds nothing)
     AND-IF (not (equal? '(? object) 'nothing))
     SAYING ("GRASP object ~a on floor." (? object))
     DELETE (goal: holds (? object))
            (holds nothing)
	    ((? object) on floor)
     ADD    (holds (? object)))
  ;;; grasp an object on the shelf
  '(grasp-on-shelf/do
    IF      (goal: holds (? object))
            (on ladder)
	    (at (? loc))
	    ((? object) at (? loc))
	    ((? object) on shelf)
	    (holds nothing)
    AND-IF  (not (equal? '(? object) 'nothing))
    SAYING  ("GRASP object ~a on shelf." (? object))
    DELETE  (goal: holds (? object))
            (goal: holds nothing)
            (holds nothing)
	    ((? object) on shelf)
    ADD     (holds (? object)))
  ;;; you can't grasp an object that isn't on the floor or a shelf
  '(grasp-invalid/do
    IF      (goal: holds (? object))
            ((? object) on (? surface))
    AND-IF  (not (or (equal? '(? object) 'nothing)
		     (equal? '(? surface) 'floor)
		     (equal? '(? surface) 'shelf)))
    SAYING ("Can't GRASP ~a since it is not on the floor or the shelf.")
    ADD    (invalid goal))
  ;;; walk to the new location to grasp an object there
  '(grasp-on-floor/walk-subgoal
    IF     (goal: holds (? object))
           ((? object) on floor)
	   ((? object) at (? new-loc))
	   (at (? old-loc))
    AND-IF (not (equal? '(? old-loc) '(? new-loc)))
    ADD    (goal: at (? new-loc)))
  ;;; if you aren't on the floor, climb down to the floor to grasp
  ;;; something that is on the floor
  '(grasp-on-floor/climb-down-subgoal
    IF     (goal: holds (? object))
           ((? object) on floor)
	   ((? object) at (? new-loc))
	   (at (? old-loc))
	   (on (? surface))
    AND-IF (not (equal? '(? surface) 'floor))
    ADD    (goal: on floor))
  ;;; move the ladder to grasp an object that is on the shelf
  '(grasp-on-shelf/move-ladder-subgoal
    IF     (goal: holds (? object))
           ((? object) on shelf)
	   ((? object) at (? new-loc))
	   (ladder at (? old-loc))
    AND-IF (not (equal? '(? old-loc) '(? new-loc)))
    ADD    (goal: ladder at (? new-loc)))
  ;;; climb the ladder to grasp an object that is on the shelf
  '(grasp-on-shelf/climb-subgoal
    IF     (goal: holds (? object))
           ((? object) on shelf)
	   ((? object) at (? loc))
	   (ladder at (? loc))
	   (on (? surface))
    AND-IF (not (equal? '(? surface) 'ladder))
    ADD    (goal: on ladder))
  ;;; get rid of whatever you are holding to grasp something new
  '(grasp/empty-hands-subgoals
    IF     (goal: holds (? object))
           (holds (? something))
	   (on (? surface))
    AND-IF (not (equal? '(? something) 'nothing))
	   (not (equal? '(? surface) 'floor))
    ADD    (goal: holds nothing))
  ;;; release an object
  '(release-something/do
    IF     (goal: holds nothing)
           (holds (? object))
	   (on (? surface))
    AND-IF (not (equal? '(? object) 'nothing))
    SAYING ("RELEASING ~a on ~a." (? object) (? surface))
    DELETE (holds (? object))
           (delete (goal: holds nothing))
    ADD    (holds nothing)
           ((? object) on (? surface)))
  ;;; release an object when you aren't holding anything
  '(release-nothing/do
    IF     (goal: holds nothing)
           (holds nothing)
    SAYING ("I'm not holding anything.")
    DELETE (goal: holds nothing))
  ;;; move an object
  '(move/do
    IF     (goal: (? object) at (? loc))
           (holds (? object))
	   (at (? loc))
    DELETE (goal: (? object) at (? loc))
    ADD    (goal: holds nothing))
  ;;; get the object in order to move it
  '(move/get-subgoal
    IF     (goal: (? object) at (? loc))
           (holds (? something))
    AND-IF (not (equal? '(? something) '(? object)))
    ADD    (goal: holds (? object)))
  ;;; walk to object you want to get
  '(move/walk-subgoal
    IF     (goal: (? object) at (? new-loc))
           (holds (? object))
	   (at (? old-loc))
    AND-IF (not (equal? '(? old-loc) '(? new-loc)))
    ADD    (goal: at (? new-loc)))
  ))
   


(define (make-full-robot-test)
  (clear-assertions)
  (remember-assertions
   '(at (9 5))
   '(on ladder)
   '(holds nothing)
   '(tools at (2 2))
   '(tools on shelf)
   '(ladder at (9 5))
   '(ladder on floor)
   '(table at (5 6))
   '(table on floor)
   '(goal: holds tools)))

(define (full-robot-test)
  (make-full-robot-rules)
  (make-full-robot-test)
  (display-assertions)
  (chain)
  (display-assertions))

