
(define (get-neighboring-places p) 
  (map car (hash-table/get neighbors p #f)))

(define (get-coordinates p) p)
  
(define (get-places p) (hash-table/get places p #f))

(define (find-intersection p1 p2)
  (let ((retval '()))
    (for-each (lambda (x)
		(for-each (lambda (y)
			    (if (equal? x y) (set! retval x)))
			  (get-places p2)))
	      (get-places p1))
    retval))
		


;;;; DEPTH-FIRST SEARCH

(define (depth-first start1 start2 finish1 finish2)
  ;;Define a queue constructor specific to depth-first search;
  ;;the depth-first-specific queue constructor adds the new paths in front:
  (define (combine-paths new-paths old-paths #!rest ignore)
    (append new-paths old-paths))
  ;;Fire up generalized search using depth-first-specific queue constructor:
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))


;;;; BREADTH-FIRST SEARCH

(define (breadth-first start1 start2 finish1 finish2)
  ;;; place new-paths in back of queue
  (define (combine-paths new-paths old-paths #!rest ignore)
    (append old-paths new-paths))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))


;;; DISTANCE MEASURE

(define (straight-line-distance start finish)
  ;;; only get the coordinates once and stick it in a variable
  (let ((sc (get-coordinates start))
	(fc (get-coordinates finish)))
    ;;; calculate the distance
    (sqrt (+ (expt (- (first-place sc) (first-place fc)) 2)
	     (expt (- (second-place sc) (second-place fc)) 2)))))

;;; CLOSER DISTANCE

(define (closer? p1 p2 finish)
  (< (straight-line-distance p1 finish) (straight-line-distance p2 finish)))

;;; BEST-FIRST SEARCH

(define (best-first start1 start2 finish1 finish2)
  ;;; take all paths and place the one that is closest to the goal
  ;;; in the front of the queue
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (closer? (first-place p1) (first-place p2) 
	       (find-intersection finish1 finish2)))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))


;;; HILL-CLIMB SEARCH

(define (hill-climb start1 start2 finish1 finish2)
  ;;; sort new-paths according to distance to goal and place them
  ;;; in front of the queue (different from best-first which sorts ALL paths)
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (closer? (first-place p1) (first-place p2) 
	       (find-intersection finish1 finish2)))
    (append  (sort new-paths sort-proc) (sort old-paths sort-proc)))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))


;;; BEAM SEARCH

(define (beam start1 start2 finish1 finish2 width)
  ;;; find ALL new paths one level deep, choose width best paths to expand
  ;;; in next sequence if necessary
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (if (= (length p1) (length p2))
	  (closer? (first-place p1) (first-place p2) 
		   (find-intersection finish1 finish2))
	  (< (length p1) (length p2))))
    (let ((combined-path (append new-paths old-paths)))
      (if (>= (length combined-path) width)
	  (list-head (sort combined-path sort-proc) width)
	  combined-path)))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))
 

;;; PATH LENGTH

(define (path-length path)
  (if (>= (length path) 2)
      (+ (straight-line-distance (car path) (cadr path))
	  (path-length (cdr path)))
      0))

;;; BRANCH-AND-BOUND SEARCH

(define (branch-and-bound start1 start2 finish1 finish2)
  ;;; add all new paths to queue, sort entire queue by path
  ;;; length with least-cost paths in front
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (< (path-length p1) (path-length p2)))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))


;;; BRANCH-AND-BOUND+ SEARCH

(define (branch-and-bound+ start1 start2 finish1 finish2)
  ;;; just like branch-and-bound except the heuristic is
  ;;; path length + estimated distance to goal
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (< (+ (path-length p1) 
	    (straight-line-distance (first-place p1) 
				    (find-intersection finish1 finish2)))
	 (+ (path-length p2) 
	    (straight-line-distance (first-place p2) 
				    (find-intersection finish1 finish2)))))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search (find-intersection start1 start2)
		      (find-intersection finish1 finish2) 
		      combine-paths))

    

(define (find-street p1 p2)
  (cadr (assoc p2 (hash-table/get neighbors p1 #f))))

(define (find-direction p1 p2)
  (if (< (abs (- (car p1) (car p2))) (abs (- (cadr p1) (cadr p2))))
      (if (> (cadr p1) (cadr p2))
          "south"
          "north")
      (if (> (car p1) (car p2))
          "west"
          "east")))


(define (describe-path p)
  (if (and (not (empty-path? p))
           (not (empty-path? (remaining-places p))))
      (let ((p1 (first-place p)) (p2 (second-place p)))
        (format #t "~%Proceed ~a on ~a for ~a units."
                (find-direction p1 p2)
                (find-street p1 p2)
                (round (straight-line-distance p1 p2)))
         (describe-path (remaining-places p)))))




;;;; GENERALIZED SEARCH PROCEDURE

(define (generalized-search start finish merge-paths-into-queue
			    #!optional queue)
  (if (default-object? queue)
      (set! queue (construct-queue (construct-path start))))
  (cond
    ;;If the queue is empty, lose; no path can be found:
    ((empty-queue? queue) #f)
    ;;If the first path in the queue terminates at the finish, win:
    ((equal? finish (first-place (first-path queue)))
     (describe-path (reverse-places (first-path queue)))
     (reverse-places (first-path queue)))
    ;;Otherwise, try again with a new, method-specific queue:
    (else (generalized-search start finish merge-paths-into-queue
            (merge-paths-into-queue (extend-path (first-path queue))
				    (remaining-paths queue)
				    finish)))))

(define (extend-path path)
  ;;Produce a list of extended paths:
  (map (lambda (new-node) (cons new-node path))
       ;;Get rid of places already visited:
       (remove-if (lambda (neighbor) (member neighbor path))
		  ;;Get the places neighboring the terminal place:
		  (get-neighboring-places (first path)))))





