;; optional-map-graphics.scm
;;
;; Uses the MIT-scheme graphics interface to provide a simple depiction of
;; the map and the current path in the progress of the search.
;; 
;; LOADING: Load this file after your streets.mdl, as it redefines
;; the GENERALIZED-SEARCH procedure, adding a few procedure calls to draw
;; the map and paths and a delay.
;;
;; EXTENT OF TESTING:  This has only been tested under X on a color 
;; decmips workstation; it is not known whether it works under Windows.
;;
;; [caroma, 1994.10.1]

;; PARAMETERS:
;;  o DELAY-PARAMETER:  An artifical delay is created by computing
;;    (delay-using-factorial delay-parameter),
;;    which requires slow bignum computations for large values.
;;    Delay-parameter defaults to 1000; you may want to reduce this for 
;;    long searches.
(define DELAY-PARAMETER 1000)

;;  o COLORS: on color screens, grey streets on black background better 
;;    allows my eye to pick out the moving bright red (active) path quickly.
;;    You may want to change these, especially on a non-color screen.
;;    E.g., make current-path-color "black" and explored-path-color "white".
(define MAP-BACKGROUND-COLOR "black")
(define MAP-COLOR "grey")
(define CURRENT-PATH-COLOR "red")
(define EXPLORED-PATH-COLOR "blue")

;; The window, initialized once when file is loaded.
(define map-window (make-graphics-device #F))

;; INTERNAL PROCEDURES used by draw-map, draw-path:

;; Selectors for coordinate pairs
(define COORDS-X first)
(define COORDS-Y second)

(define (FIND-COORDINATE-EXTREMES continuation)
  ;; Find the minimum and maximum x and y coordinate values of all the
  ;; places in the neighbors table.  (Data may not always have
  ;; coordinates table, e.g., cambridge streets data)
  (define (loop keys xmin xmax ymin ymax)
    (if (Null? keys)
	;; if no more keys, done; pass 4 values to continuation procedure
	(continuation xmin xmax ymin ymax)
	;; else try next coord pair and recurse with new mins and maxes.
	(let ((next-coords (Get-coordinates (first keys))))
	  (loop (Cdr keys)
		(min xmin (Coords-x next-coords))
		(max xmax (Coords-x next-coords))
		(min ymin (Coords-y next-coords))
		(max ymax (Coords-y next-coords))))))
  (let ((keys (Hash-table/key-list neighbors)))
    (if (Null? keys)
	(Error "No neighbors in hash table.")
	(let ((key1-coords (Get-coordinates (First keys))))
	  ;; use first place coords as initial values for xmin,xmax,ymin,ymax
	  (loop (Cdr keys)
		(Coords-x key1-coords) (Coords-x key1-coords)
		(Coords-y key1-coords) (Coords-y key1-coords))))))
	  
(define (SET-GRAPH-COORDINATES-FOR-MAP)
  ;; Compute virtual coordinates so that map fits in window.
  ;; Size is square so that map is not distorted in square window.
  (Find-coordinate-extremes
   (lambda (xmin xmax ymin ymax)
     ;; default window is square, so find max difference
     (let* ((maxdiff (Max (- xmax xmin) (- ymax ymin)))
	    ;; add a little for a margin, divide by 2
	    (maxradius (/ (* 1.2 maxdiff) 2))
	    (xcenter (/ (+ xmin xmax) 2))
	    (ycenter (/ (+ ymin ymax) 2)))
       (Graphics-set-coordinate-limits
	map-window
	(- xcenter maxradius) (+ ycenter maxradius)
	(+ xcenter maxradius) (- ycenter maxradius))))))
       
(define (DRAW-MAP)
  ;; Set colors and virtual coordinates, clear window, and draw map.
  ;; For each place, draw name and connections to previously drawn places.
  (Graphics-operation Map-window 'set-background-color Map-background-color)
  (Graphics-operation Map-window 'set-foreground-color Map-color)
  (Set-graph-coordinates-for-map)
  (Graphics-clear Map-window)
  (let ((places (Hash-table/key-list Neighbors)))
    (define (DRAW-PLACE place)
      (let ((coords (Get-coordinates place))
	    (undrawn-neighbors (Cdr (Member place places))))
       (define (DRAW-CONNECTION neighboring-place)
	 (if (Member neighboring-place undrawn-neighbors)
	     ;; avoid drawing lines twice
	     #F
	     (let ((neighbor-coords (Get-coordinates neighboring-place)))
	       (Graphics-draw-line
		map-window
		(Coords-x coords)          (Coords-y coords)
		(Coords-x neighbor-coords) (Coords-y neighbor-coords)))))
       (For-each draw-connection (Get-neighboring-places place))
       (Graphics-draw-text Map-window
			   (Coords-x coords) (Coords-y coords)
			   (Format #F " ~a" place))))
    (For-each draw-place places)))
			   
(define (DRAW-PATH path #!optional color)
  ;; Set color and draw path.
  ;; Draw path by drawing each segment in path.
  (define (LOOP path)
    (if (Not (Null? (Remaining-places path)))
	(let ((from-coords (Get-coordinates (First-place path)))
	      (to-coords   (Get-coordinates (Second-place path))))
	  (Graphics-draw-line Map-window
			      (Coords-x from-coords) (Coords-y from-coords)
			      (Coords-x to-coords)   (Coords-y to-coords))
	  (loop (Remaining-places path)))))
  (if (Not (Default-object? color))
      (Graphics-operation Map-window 'set-foreground-color color))
  (loop path))


;;; delay
(define (DELAY-USING-FACTORIAL n)
  (do ((p 1)
       (i n (- i 1)))
      ((<= i 0) p)
    (set! p (* p i))))


;;;; GENERALIZED SEARCH PROCEDURE  -- orginally in streets.mdl

(define (GENERALIZED-SEARCH start finish merge-paths-into-queue
			    #!optional queue)
  (if (Default-object? queue)  ;; on first call, initialize queue, map
      (begin (set! queue (Construct-queue (Construct-path start)))
	     (DRAW-MAP)))
  (if (Not (Empty-queue? queue))
      ;; path is reversed here to draw it in the forward direction;
      ;; below it is erased in the reverse direction for best visual effect
      (begin (DRAW-PATH (Reverse-places (First-path queue)) Current-path-color)
	     (DELAY-using-factorial delay-parameter)))
  (cond
    ;;If the queue is empty, lose; no path can be found:
    ((Empty-queue? queue) #f)
    ;;If the first path in the queue terminates at the finish, win:
    ((Equal? finish (First-place (First-path queue)))
     (Format #t
	     "~%The successful path is ~a."
	     (Reverse-places (First-path queue)))
     (Reverse-places (First-path queue)))
    ;;Otherwise, try again with a new, method-specific queue:
    (else (DRAW-PATH (First-path queue) Explored-path-color)
	  (Generalized-search start finish merge-paths-into-queue
            (merge-paths-into-queue (Extend-path (First-path queue))
				    (Remaining-paths queue)
				    finish)))))
