;;;; DEPTH-FIRST SEARCH

(define (depth-first start finish)
  ;;Define a queue constructor specific to depth-first search;
  ;;the depth-first-specific queue constructor adds the new paths in front:
  (define (combine-paths new-paths old-paths #!rest ignore)
    (append new-paths old-paths))
  ;;Fire up generalized search using depth-first-specific queue constructor:
  (generalized-search start finish combine-paths))


;;;; BREADTH-FIRST SEARCH

(define (breadth-first start finish)
  ;;; place new-paths in back of queue
  (define (combine-paths new-paths old-paths #!rest ignore)
    (append old-paths new-paths))
  (generalized-search start finish combine-paths))


;;; DISTANCE MEASURE

(define (straight-line-distance start finish)
  ;;; only get the coordinates once and stick it in a variable
  (let ((sc (get-coordinates start))
	(fc (get-coordinates finish)))
    ;;; calculate the distance
    (sqrt (+ (expt (- (first-place sc) (first-place fc)) 2)
	     (expt (- (second-place sc) (second-place fc)) 2)))))

;;; CLOSER DISTANCE

(define (closer? p1 p2 finish)
  (< (straight-line-distance p1 finish) (straight-line-distance p2 finish)))

;;; BEST-FIRST SEARCH

(define (best-first start finish)
  ;;; take all paths and place the one that is closest to the goal
  ;;; in the front of the queue
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (closer? (first-place p1) (first-place p2) finish))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search start finish combine-paths))


;;; HILL-CLIMB SEARCH

(define (hill-climb start finish)
  ;;; sort new-paths according to distance to goal and place them
  ;;; in front of the queue (different from best-first which sorts ALL paths)
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (closer? (first-place p1) (first-place p2) finish))
    (append  (sort new-paths sort-proc) (sort old-paths sort-proc)))
  (generalized-search start finish combine-paths))


;;; BEAM SEARCH

(define (beam start finish width)
  ;;; find ALL new paths one level deep, choose width best paths to expand
  ;;; in next sequence if necessary
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (if (= (length p1) (length p2))
	  (closer? (first-place p1) (first-place p2) finish)
	  (< (length p1) (length p2))))
    (let ((combined-path (append new-paths old-paths)))
      (if (>= (length combined-path) width)
	  (list-head (sort combined-path sort-proc) width)
	  combined-path)))
  (generalized-search start finish combine-paths))
 

;;; PATH LENGTH

(define (path-length path)
  (if (>= (length path) 2)
      (+ (straight-line-distance (car path) (cadr path))
	  (path-length (cdr path)))
      0))

;;; BRANCH-AND-BOUND SEARCH

(define (branch-and-bound start finish)
  ;;; add all new paths to queue, sort entire queue by path
  ;;; length with least-cost paths in front
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (< (path-length p1) (path-length p2)))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search start finish combine-paths))


;;; BRANCH-AND-BOUND+ SEARCH

(define (branch-and-bound+ start finish)
  ;;; just like branch-and-bound except the heuristic is
  ;;; path length + estimated distance to goal
  (define (combine-paths new-paths old-paths #!rest ignore)
    (define (sort-proc p1 p2)
      (< (+ (path-length p1) (straight-line-distance (first-place p1) finish))
	 (+ (path-length p2) (straight-line-distance (first-place p2) 
						     finish))))
    (sort (append new-paths old-paths) sort-proc))
  (generalized-search start finish combine-paths))

