;;;; -*- mode:Scheme -*- ;;;;

;;;; AUXILIARIES FOR COMPUTING DISTANCES

(define *earth-diameter* 7926.0)

(define *pi* (* 4 (atan 1 1)))

(define (city-lat city) (first (hash-table/get *places* city #f)))

(define (city-long city) (second (hash-table/get *places* city #f)))

(define (air-distance city1 city2)
  "The great circle distance between two cities."
  (let ((d (distance (xyz-coords city1) (xyz-coords city2))))
    ;; d is the straight-line chord between the two cities,
    ;; The length of the subtending arc is given by:
    (* *earth-diameter* (asin (/ d 2)))))

(define (xyz-coords city)
  "Returns the x,y,z coordinates of a point on a sphere.
  The center is (0 0 0) and the north pole is (0 0 1)."
  (let ((psi (deg->radians (city-lat city)))
        (phi (deg->radians (city-long city))))
    (list (* (cos psi) (cos phi))
          (* (cos psi) (sin phi))
          (sin psi))))

(define (distance point1 point2)
  "The Euclidean distance between two points.
  The points are coordinates in n-dimensional space."
  (sqrt (reduce + 0 (map (lambda (a b) (expt (- a b) 2))
			 point1 point2))))

(define (deg->radians deg)
  "Convert degrees and minutes to radians."
  (* (+ (truncate deg) (* (- deg (truncate deg)) 100/60)) *pi* 1/180))


;;;; OTHER AUXILIARY PROCEDURES

(define (path-length path)
  (if (null? (cdr path))
      0
      (+ (air-distance (first path) (second path))
         (path-length (cdr path)))))

(define (flight-time flight) (/ (path-length flight) *mph*))

(define (ready-to-fly-time flight)
  (+ (first flight)
     (flight-time (cdr flight))
     (* (- (length flight) 2) *ground-time*)))

(define (valid-fleet-assignment? assignment)
  (cond ((null? assignment) #t)
	((valid-airplane-assignment? (cdr (first assignment)))
	 (valid-fleet-assignment? (cdr assignment)))
	(else #f)))

(define (valid-airplane-assignment? assignment)
  (cond ((null? assignment) #t)
	((null? (cdr assignment)) #t)
	((let ((first-leg (first assignment))
	       (second-leg (second assignment)))
	   (>= (first second-leg)
	       (+ (ready-to-fly-time first-leg)
		  (let ((place1 (first (last-pair first-leg)))
			(place2 (second second-leg)))
		    (if (eq? place1 place2)
			0
			(+ *ground-time*
			   (flight-time (list place1 place2))))))))
	 (valid-airplane-assignment? (cdr assignment)))
	(else #f)))

(define (add-assignment aircraft flight a-list)
  (cond ((null? a-list)
	 (list (list aircraft flight)))
	((eq? aircraft (first (first a-list)))
	 (cons (cons aircraft
		     (sort (cons flight (cdr (first a-list)))
			   (lambda (x y) (< (first x) (first y)))))
	       (cdr a-list)))
	(else (cons (first a-list)
		    (add-assignment aircraft flight (cdr a-list))))))


