;;;; -*- mode:Scheme -*- ;;;;

(declare (usual-integrations))

(load-option 'format)

;;;; GLOBAL VARIABLES

(define *training-data*)

(define *test-data*)

(define *tree*)

(define *normalization-vector*)

;;;; COMMONLISP IMITATOR

(define (remove-duplicates l)
  (cond ((null? l) ())
	((member (first l) (cdr l)) (remove-duplicates (cdr l)))
	(else (cons (first l) (remove-duplicates (cdr l))))))

;;;; STRUCTURE AND STRUCTURE ACCESSORS

(define node
  (make-record-type "node"
    '(dimension			;The dimension that the node divides on.
      gap-min			;The maximum of the "small" samples.
      gap-max			;The minimum of the "large" samples.
      left-branch	        ;The left sample(s) or node.
      right-branch	        ;The right sample(s) or node.
      )))

(define make-node (record-constructor node))

(define node? (record-predicate node))

(define node-dimension (record-accessor node 'dimension))
(define node-gap-min (record-accessor node 'gap-min))
(define node-gap-max (record-accessor node 'gap-max))

(define node-left-branch (record-accessor node 'left-branch))
(define node-right-branch (record-accessor node 'right-branch))

(define set-node-left-branch (record-modifier node 'left-branch))
(define set-node-right-branch (record-modifier node 'right-branch))

;;; Access Functions for Records Stored in List of Samples

(define (record-diagnosis record) (first record))

(define (record-attributes record) `(,(list-ref record 2) 
				     ,(list-ref record 9) 
				     ,@(list-tail record 16)))

;;;; CONSTRUCT TREE

(define *maximum-depth* 0)

(define (make-tree)
  "
  Purpose:	Supply arguments to make-tree-aux and assign value to *tree*.
  Returns:	A tree node (a structure).
  "
  (set! *maximum-depth* 0)
  (set! *tree* (make-tree-aux *training-data* 0))
  (format #t "~%The maximum depth of the tree is ~a" *maximum-depth*)
  #t)

(define (make-tree-aux samples level)
  "
  Purpose:	Construct a KD tree from samples.
  Returns:	A tree node (a structure).
  "
  (set! *maximum-depth* (max *maximum-depth* level))
  (let* ((dimension-count (length (record-attributes (first samples))))
	 (dimension-to-check (remainder level dimension-count)))
    (cond
     ((test-for-variation dimension-to-check samples)
      (if (not (cdr samples))
	  (first samples)
	  (let ((node (make-node-for-samples dimension-to-check samples)))
	    ;;Recurse along both branches:
	    (set-node-left-branch
	     node
	     (make-tree-aux (node-left-branch node) (1+ level)))
	    (set-node-right-branch
	     node
	     (make-tree-aux (node-right-branch node) (1+ level)))
	    node)))
     (else (say *verbose*
		"~%Skipping dimension ~a on level ~a; no variation"
		dimension-to-check level)
	   (make-tree-aux samples (1+ level))))))

(define (test-for-variation dimension samples)
  "
  Purpose:	See if the samples vary in the given dimension.
  Returns:	#f or #t
  "
  (let ((reference (list-ref (record-attributes (first samples)) dimension)))
    (define (loop others)
      (if (null? others)
	  #f
	  (if (equal? reference
		      (list-ref (record-attributes (first others)) dimension))
	      (loop (cdr others))
	      #t)))
    (if (null? (cdr samples))
	#t
	(loop (cdr samples)))))

(define (make-node-for-samples dimension samples)
  "
  Purpose:	Sorts samples along dimension supplied and creates a node.
  Returns:	A node.
  "
  (let* (;; Sort the samples:
	 (samples
	  (sort samples
		(lambda (x y)
		  (< (list-ref (record-attributes x) dimension)
		     (list-ref (record-attributes y) dimension)))))
	 ;; Extract numbers in given dimension:
	 (numbers (map (lambda (x) (list-ref (record-attributes x) dimension))
		       samples))
	 ;; Compute the position of the rightmost element of the list:
	 (index-limit (-1+ (length numbers)))
	 ;; Compute the position of the center of the list:
	 (left-size (quotient (length numbers) 2)))
    (define (seek-nearest-split n)
      "
      Purpose:	Finds nearest pair of samples that differ in the dimension.
                Sets left-size variable accordingly.
      "
      (let* ((ll (max 0 (-1+ (- left-size n))))
	     (lr (1+ ll))
	     (rr (min index-limit (+ left-size n)))
	     (rl (-1+ rr)))
	(if (or (not (<= 0 ll index-limit))
		(not (<= 0 lr index-limit))
		(not (<= 0 rl index-limit))
		(not (<= 0 rr index-limit)))
	    (begin
	      (format #t "~%Limit problem:")
	      (pp (list ll lr rl rr))
	      (pp numbers)
	      (error "")))
	(cond ((not (= (list-ref numbers ll) (list-ref numbers lr)))
	       (set! left-size (1+ ll)))
	      ((not (= (list-ref numbers rl) (list-ref numbers rr)))
	       (set! left-size (1+ rl)))
	      (else (seek-nearest-split (1+ n))))))
    ;; Reset left-size, if necessary:
    (seek-nearest-split 0)
    ;; Split up samples and make a node:
    (make-node
     dimension
     (list-ref numbers (-1+ left-size))
     (list-ref numbers left-size)
     (list-head samples left-size)
     (list-tail samples left-size))))

;;;; DISPLAY TREE

(define (show-tree)
  "
  Purpose:	Supply argument to show-tree-aux.
  "
  (show-tree-aux *tree* 0 #f))

(define (show-tree-aux node level branch)
  "
  Purpose:	Display a tree using indentation to indicate level.
  Arguments:	The root node of the tree.
  "
  (if (node? node)
      (format #t "~%~a~a split on dimension ~a [~a ~a]"
	      (indent level)
	      (if branch branch "Top")
	      (node-dimension node)
	      (node-gap-min node)
	      (node-gap-max node))
      (format #t "~%~a~a" (indent level) node))
  (if (node? node)
      (begin
	(show-tree-aux
	 (node-left-branch node) (1+ level) "Left branch")
	(show-tree-aux
	 (node-right-branch node) (1+ level) "Right branch"))))

(define (indent depth) (make-string depth #\space))

;;;; NUMERICAL AUXILIARIES

(define (distance2 u v)
  "
  Purpose:	Computes squared weighted distance between two vectors.
  "
  (reduce + 0
	  (map (lambda (x y z) (delta2 x y z))
	       u v *normalization-vector*)))

(define (delta2 x1 x2 n)
  "
  Purpose:	Computes squared weighted distance between two points.
  "
  (let ((delta (- x1 x2)))
    (* delta delta n)))

;;;; DATA READERS

(define (read-training-data file)
  "
  Purpose:	Read training data.  Assign *training-data*.
  "
  (set! *training-data* (read-data file))
  (make-all-ones-vector)
  #t)

(define (read-test-data file)
  "
  Purpose:	Read test data.  Assign *test-data*.
  "
  (set! *test-data* (read-data file))
  #t)

(define (read-data file)
  "
  Purpose:	Read list of expressions from a file.
  Remark:       Complains if expressions vary in length.
  Returns:	A list of expressions.
  "
  (let ((result '())
	(record-length 0)
	(count 0))
    (define (read-expression)
      (let ((e (read)))
	(if (eof-object? e)
	    'done
	    (begin
	      (if (zero? record-length)
		  (set! record-length (length e))
		  (if (not (= (length e) record-length))
		      (format #t "~%Record ~a is wrong length" count)))
	      (set! result (cons e result))
	      (set! count (1+ count))
	      (read-expression)))))
    (with-input-from-file file read-expression)
    result))

;;;; NORMALIZER

(define (average numbers)
  "
  Purpose:	Find average of a list of numbers.
  "
  (/ (reduce + 0.0 numbers) (length numbers)))

(define (average-of-squared-deviations numbers)
  "
  Purpose:	Find square root of average of
                squares of deviations of numbers from their average.
  "
  (let ((a (average numbers)))
    (define (square n) (* n n))
    (average (map (lambda (n) (square (- n a))) numbers))))

(define (make-all-ones-vector)
  "
  Purpose:	Make vector with 1s in each dimension.
  Remark:       Assigns *normalization-vector*
  "
  (set! *normalization-vector*
	(map (lambda (x) 1) (record-attributes (first *training-data*)))))

(define (make-normalizing-vector)
  "
  Purpose:	Make normalizing vector with variance equalizing number
                in each dimension.
  Remark:       Assigns *normalization-vector*
  "
  (define (loop samples)
    (if (null? (first samples))
	'()
	(cons (average-of-squared-deviations (map first samples))
	      (loop (map cdr samples)))))
  (set! *normalization-vector*
	(map (lambda (x) (/ 1.0 x))
	     (loop (map record-attributes *training-data*))))
  *normalization-vector*)


;;;; IDENTIFICATION

(define (weight-evenly x)
  "
  Purpose:	Weight neighbors evenly no matter what the distance.
  "
  1)

(define (weight-by-inverse-distance x)
  "
  Purpose:	Weight neighbors evenly no matter what the distance.
  "
  (/ (sqrt x)))

(define *neighbors* 1)

(define *weighting-function* weight-evenly)

(define (identify unknown)
  (identify-aux unknown *tree* *neighbors* *weighting-function*))

(define (identify-aux unknown tree #!optional count weighting-function)
  "
  Purpose:	To guess an attribute using nearest neighbor idea.
  Arguments:	unknown: a list of dimension values
		tree: a KD tree.
		count: the number of nearest neighbors to be used.
		weighting-function: a function that establishes the way
				    a neighbor's influence is diminished
				    by distance.
  Returns:	The best guess for an unknown's diagnosis.
  "
  (let ((candidates (find-answers unknown tree count)))
    (format #t "~%Unknown:")
    (newline)
    (display unknown)
    (format #t "~%Candidate(s):")
    (for-each (lambda (x y)
		(newline) (display x)
		(display " ") (display (sqrt y)))
	      (map answer-record candidates)
	      (map answer-distance2 candidates))
    (if (= 0 (answer-distance2 (first candidates)))
	;;If the answer is at the same place exactly, report it:
	(let ((winner (answer-diagnosis (first candidates))))
	  (format #t "~%The winner is ~a (exact match)." winner)
	  winner)
      ;;Otherwise, tally up votes:
      (let* ((best-pair (tally-votes candidates weighting-function))
	     (winner (first best-pair))
	     (score (second best-pair)))
	(format #t "~%The winner is ~a with ~a votes." winner score)
	winner))))

(define (tally-votes attribute-distance-pairs weighting-function)
  "
  Purpose: Combines evidence when there are multiple nearest neighbors.
  Returns: Attribute with the most votes.
  "
  (let* ((attribute-weight-pairs
	  ;;Make a-list pairs in which first element is an object's attribute 
	  ;;and the second element is the distance-determined influence of
	  ;;that object attribute.
	  (map (lambda (e)
		 (list (answer-diagnosis e)
		       (weighting-function (answer-distance2 e))))
	       attribute-distance-pairs))
	 (attribute-score-pairs 
	  ;;Make a-list pairs in which first element is an attribute and the
	  ;;second element is the sum of the influences of the objects
	  ;;with that attribute:
	  (map (lambda (attribute)
		 (list attribute
		       (accumulate-weight attribute attribute-weight-pairs)))
	       (remove-duplicates (map first attribute-weight-pairs)))))
    ;;Sort, with the most recommended attribute in front:
    (set! attribute-score-pairs
	  (sort attribute-score-pairs
		(lambda (x y) (> (second x) (second y)))))
    (format #t "~%The scores are: ~a" attribute-score-pairs)
    ;;Pick the winner off the front:
    (first attribute-score-pairs)))

(define (accumulate-weight attribute pairs)
  "
  Purpose:	Helper
  "
  (cond ((null? pairs) 0)
	((eq? attribute (first (first pairs)))
	 (+ (second (first pairs)) (accumulate-weight attribute (cdr pairs))))
	(else (accumulate-weight attribute (cdr pairs)))))

;;;; FIND NEAREST NEIGHBORS

(define (find-answers unknown tree #!optional count level)
  "
  Purpose:	Find N nearest neighbors using KD tree.
  Returns:	N answers
  Remarks:	This procedure is complicated, in part, because
		it has to deal with multiple nearest neighbors.
  "
  (if (default-object? count) (set! count 1))
  (if (default-object? level) (set! level 0))
  (let* ((dimension (node-dimension tree))
	 (projection (list-ref (record-attributes unknown) dimension))
	 (left-delta2 (delta2 projection
			      (node-gap-min tree)
			      (list-ref *normalization-vector* dimension)))
	 (right-delta2 (delta2 projection
			       (node-gap-max tree)
			       (list-ref *normalization-vector* dimension)))
	 (threshold-delta2) (winning-branch) (losing-branch))
    ;;Decide which branch has won and set variables accordingly:
    (if (< right-delta2 left-delta2)
	(begin
	  (say *verbose* "~%~aTurn toward large numbers at level ~a(~a): ~
		       ~a is closer to ~a than to ~a."
		       (indent level)
		       level dimension projection
		       (node-gap-max tree)
		       (node-gap-min tree))
	       (set! threshold-delta2 left-delta2)
	       (set! winning-branch (node-right-branch tree))
	       (set! losing-branch (node-left-branch tree)))
      (begin (say *verbose* "~%~aTurn toward small numbers at level ~a(~a): ~
		     ~a is closer to ~a than to ~a."
		     (indent level)
		     level dimension projection 
		     (node-gap-min tree)
		     (node-gap-max tree))
	     (set! threshold-delta2 right-delta2)
	     (set! winning-branch (node-left-branch tree))
	     (set! losing-branch (node-right-branch tree))))
    ;;At this point, it looks like the winning direction is known.
    ;;This may prove wrong later, of course, because the decision
    ;;is based on comparison in one dimension only, not on actual distance.
    (let* ((winning-answers
	    ;;If the winning branch is a node, then find the
	    ;;closest neighbor by calling find-answers recursively;
	    ;;otherwise the winning branch is not a node, and the winning
	    ;;branch is the closest neighbor:
 	    (if (node? winning-branch)
		(find-answers unknown winning-branch count (1+ level))
		(list
		 (make-answer
		  winning-branch
		  (distance2 (record-attributes unknown)
			     (record-attributes winning-branch))))))
	   (losing-answers)
	   (nearest-winning-distance2
	    (answer-distance2 (first (last-pair winning-answers)))))
      ;;at this point, find-answers needs to check the nearest answer
      ;;by comparing the actual distance between the unknown and
      ;;the nearest answer with the one-dimensional distance between
      ;;the unknown and the nearest answer on the wrong side of
      ;;the gap between the left and right groups; also, there
      ;;may not yet be enough answers:
      (if (and (<= nearest-winning-distance2 threshold-delta2)
	       (>= (length winning-answers) count))
	  ;;If the answer holds up, done:
	  winning-answers
	;;Otherwise, find the best answers on the other side too:
	(begin
	  ;;Indicate why there is more work to do:
	  (if (<= nearest-winning-distance2 threshold-delta2)
	      (say
	       *verbose*
	       "~%~aTrying alternate branch because too few answers ~
                  [~a <= ~a]."
	       (indent level)
	       (length winning-answers) count)
	    (say *verbose* "~%~aTrying other branch at level ~a ~
                              because worst answer ~
 		              is not good enough [~a > ~a]."
		 (indent level) level
		 nearest-winning-distance2 threshold-delta2))
	  ;;Establish best answers on the losing branch of the tree:
	  (set! losing-answers
		(if (node? losing-branch)
		    (find-answers unknown losing-branch count (1+ level))
		  (list (make-answer
			 losing-branch
			 (distance2 (record-attributes unknown)
				    (record-attributes losing-branch))))))
	  ;;Combine all answers and sort:
	  (set! winning-answers
		(sort (append winning-answers losing-answers)
		      (lambda (x y) (< (answer-distance2 x)
				       (answer-distance2 y)))))
	  ;;Prune if too many answers:
	  (if (> (length winning-answers) count)
	      (set! winning-answers (list-head winning-answers count)))
	  ;;Finally, return the answers:
	  winning-answers)))))

(define (make-answer record d2)
  (list (record-diagnosis record) d2 record))
(define (answer-diagnosis answer) (first answer))
(define (answer-distance2 answer) (second answer))
(define (answer-record answer) (third answer))

;;;; DEBUGGING

(define (say switch #!rest args)
  "
  Purpose:	Display stuff if switch is set to #t.
  "
  (if switch (apply format #t args)))

(define *verbose* #f)

