;;;; -*- mode:Scheme -*- ;;;;

;;;; DECLARATION FOR THE COMPILER

(declare (usual-integrations))

;;;; USE THE FORMAT LIBRARY

(load-option 'format)

;;;; GLOBAL VARIABLES

(define *training-data*)    ; Assigned to a list of training records

(define *test-data*)        ; Assigned to a list of test records

(define *tree*)             ; Assigned by MAKE-TREE to an identification tree

;;;; COMMONLISP IMITATORS

(define (remove-duplicates l)
  "
  Purpose:	   Remove duplicates in a list.
  Sample argument: (a b c b d)
  Sample value:    (a b c d)
  "
  (cond ((null? l) ())
	((member (first l) (cdr l)) (remove-duplicates (cdr l)))
	(else (cons (first l) (remove-duplicates (cdr l))))))

(define (count x l)
  "
  Purpose:	   Counts the number of times a symbol appears in a list.
  Sample argument: (a b c b a)
  Sample value:    2
  "
  (let ((n 0))
    (for-each (lambda (e) (if (eq? e x) (set! n (1+ n))))
	      l)
    n))

;;;; STRUCTURE AND STRUCTURE ACCESSORS

;;; A structure type for non-terminal tree nodes:

(define branching-node
  (make-record-type "node"
    '(dimension			;The branching dimension.
      threshold			;The dividing line.
      left-branch	        ;The left node
      right-branch	        ;The right node.
      )))

;;; Structure Constructors, Predicates, and Getters:

(define make-branching-node (record-constructor branching-node))

(define branching-node? (record-predicate branching-node))

(define branching-node-dimension
  (record-accessor branching-node 'dimension))
(define branching-node-threshold
  (record-accessor branching-node 'threshold))
(define branching-node-left-branch
  (record-accessor branching-node 'left-branch))
(define branching-node-right-branch
  (record-accessor branching-node 'right-branch))

;;; A structure type for terminal tree nodes

(define terminal-node
  (make-record-type "node"
    '(samples
      identity
      )))

;;; Structure Constructors, Predicates, and Getters:

(define make-terminal-node (record-constructor terminal-node))

(define terminal-node? (record-predicate terminal-node))

(define terminal-node-samples
  (record-accessor terminal-node 'samples))

(define terminal-node-identity
  (record-accessor terminal-node 'identity))

;;;; ACCESS FUNCTIONS FOR TRAINING AND TEST RECORDS

(define (record-diagnosis record) (first record))

(define (record-attributes record) (cdr record))

;;;; TREE CONSTRUCTOR

(define *maximum-depth*)		; A global used by MAKE-TREE and
					; MAKE-TREE-AUX

(define (make-tree)
  "
  Purpose:	Supplies arguments to MAKE-TREE-AUX
                Assigns value to *tree*.
                Initializes *maximum-depth*.
  Returns:	A branching node (a structure).
  "
  (set! *maximum-depth* 0)
  (set! *tree* (make-tree-aux *training-data* 0))
  (format #t "~%The maximum depth of the tree is ~a" *maximum-depth*)
  #t)

(define (make-tree-aux samples level)
  "
  Purpose:	Construct an ID3 tree from samples.
  Returns:	A branching node (a structure).
  "
  ;; Keep track of the maximum depth of recursion, which equals the tree depth:
  (set! *maximum-depth* (max *maximum-depth* level))
  (let ((dimension-count (length (record-attributes (first samples))))
	(disorder 10000)
	(dimension 0)
	(threshold 0)
	(left-samples '())
	(right-samples '()))
    ;; Find the dimension and threshold that produces the minimum disorder.
    ;; First, try each dimension:
    (for-each
     (lambda (d)
       ;; During which, try each threshold:
       (for-each
	(lambda (t)
	  (let* ((answer (score-threshold t d samples))
		 (score (answer-score answer)))
	    ;; If the current score is lower than the best so far,
	    ;; then reset variables accordingly:
	    (if (< score disorder)
		(begin (set! disorder score)
		       (set! dimension d)
		       (set! threshold t)
		       (set! left-samples (answer-left-samples answer))
		       (set! right-samples (answer-right-samples answer))
		       ))))
	;; Supply a list of thresholds:
	(compute-thresholds-in-dimension d samples)))
     ;; Supply a list of dimensions:
     (list-dimensions dimension-count))
    ;; At this point, announce the disorder minimizing dimension and threshold:
    (format #t "~%Splitting in dimension ~a at threshold ~a ..."
	    dimension threshold)
    ;; Make a new node, using the discovered dimension and threshold:
    (make-branching-node
     dimension
     threshold
     ;; Also, if the remaining samples in the left branch suggest varying
     ;; diagnoses ...
     (if (test-for-variation left-samples)
	 ;; then make another branching node on the left side,
	 (make-tree-aux left-samples (1+ level))
	 ;; otherwise, make a terminal node:
	 (make-terminal-node left-samples
			     (record-diagnosis (first left-samples))))
     ;; Repeat for the right branch:
     (if (test-for-variation right-samples)
	 (make-tree-aux right-samples (1+ level))
	 (make-terminal-node right-samples
			     (record-diagnosis (first right-samples)))))))

(define (test-for-variation samples)
  "
  Purpose:	See if the samples vary:
  Returns:	#t or #f
  "
  (let ((reference (record-diagnosis (first samples))))
    ;; Define a testing loop:
    (define (loop others)
      (if (null? others)
	  ;; If no more samples to test, there is no variation:
	  #f
	  ;; If there is at least one, compare with reference;
	  ;; If same, keep going; if not, there is variation:
	  (if (equal? reference (record-diagnosis (first others)))
	      (loop (cdr others))
	      #t)))
    ;; If there is just one sample ...
    (if (null? (cdr samples))
	;; there is no variation:
	#f
	;; otherwise, launch the testing loop:
	(loop (cdr samples)))))

(define (list-dimensions count)
  "
  Purpose:	   Given a number of dimensions, return a list:
  Sample argument: 4
  Sample value:	   (0 1 2 3)
  "
  (define (loop n)
    (if (= n count)
	'()
	(cons n (loop (1+ n)))))
  (loop 0))

(define (compute-thresholds numbers)
  "
  Purpose:	   Given a list of numbers, return all intermediate numbers:
  Sample argument: (1 2 2 3 3 4 5)
  Sample value:	   (1.5 2.5 3.5 5.)
  "
  (let ((numbers (remove-duplicates (sort numbers <))))
    (map (lambda (x y) (* 0.5 (+ x y))) ; Compute average
	 numbers			; The numbers, sans duplicates
	 (cdr numbers))))		; The numbers, sans duplicates, shifted

(define (compute-thresholds-in-dimension dimension samples)
  "
  Purpose:	Find potential thresholds
  Returns:      A list of thresholds
  Remark:       COMPUTE-THRESHOLDS does the work
  "
  (compute-thresholds
   (map (lambda (x) (list-ref (record-attributes x) dimension))
	samples)))

(define (score-threshold threshold dimension samples)
  "
  Purpose:	Compute disorder score for given dimension and threshold.
  Returns:	A list of the score and two lists into which the samples
                are divided by the threshold in the dimension
  "
  (let ((left-samples '())
	(right-samples '())
	(left-count 0)
	(right-count 0))
    ;; Loop over the samples:
    (for-each
     (lambda (x)
       ;; If the appropriate attribute value is less than the threshold ...
       (if (< (list-ref (record-attributes x) dimension)
	      threshold)
	   ;; Increment information describing the left branch:
	   (begin
	     (set! left-samples (cons x left-samples))
	     (set! left-count (1+ left-count)))
	   ;; Otherwise, increment information describing the right branch:
	   (begin
	     (set! right-samples (cons x right-samples))
	     (set! right-count (1+ right-count)))))
     samples)
    ;; Given the way the threshold divides the samples ...
    (let ((total-count (+ left-count right-count)))
      (package-score
       ;; Return the disorder score:
       (+ (* (/ left-count total-count)
	     (disorder-of-set left-count (map first left-samples)))
	  (* (/ right-count total-count)
	     (disorder-of-set right-count (map first right-samples))))
       ;; Along with the way the samples were divided:
       left-samples
       right-samples))))

(define package-score list)		; Constructor

(define answer-score first)		; Getters

(define answer-left-samples second)

(define answer-right-samples third)

;;;; DISORDER CALCULATIONS

(define (disorder-of-set size set)
  "
  Purpose: Compute disorder of a set
  Returns: Disorder value
  "
  (reduce + 0 (map (lambda (x) (compute-disorder-term x size))
		   (map (lambda (x) (count x set))
			(remove-duplicates set)))))


(define (compute-disorder-term x total)
  (let ((ratio (/ x total)))
    (- (* ratio (log2 ratio)))))

(define (log2 x) (/ (log x) (log 2)))


;;;; DISPLAY TREE

(define (show-tree)
  "
  Purpose:	Supply arguments to show-tree-aux.
  "
  (show-tree-aux *tree* 0 #f))

(define (show-tree-aux node level branch)
  "
  Purpose:	Display a tree using indentation to indicate level
  Arguments:	The root node of the tree, the level, and the direction
  "
  (cond ((branching-node? node)
	 ;; It's a branching node; describe it:
	 (format #t "~%~a~a split on dimension ~a [at ~a]"
		 (indent level)
		 ;; Usually BRANCH's value is a string, but not
		 ;; when SHOW-TREE-AUX is called by SHOW-TREE:
		 (if branch branch "Top")
		 (branching-node-dimension node)
		 (branching-node-threshold node))
	 ;; Recurse left:
	 (show-tree-aux
	  (branching-node-left-branch node) (1+ level) "Left branch")
	 ;; Recurse right:
	 (show-tree-aux
	  (branching-node-right-branch node) (1+ level) "Right branch"))
	(else
	 ;; It is a terminal node; describe it:
	 (let ((count (length (terminal-node-samples node))))
	   (format #t "~%~aAnswer: ~a [~a sample~a]"
		   (indent level)
		   (terminal-node-identity node)
		   count
		   ;; Get the english correct:
		   (if (> count 1) "s" ""))))
	))

(define (indent depth) (make-string depth #\space))

;;;; DATA READERS

(define (read-training-data file)
  "
  Purpose:	Read training data.  Assign *training-data*.
  "
  (set! *training-data* (read-data file))
  #t)

(define (read-test-data file)
  "
  Purpose:	Read test data.  Assign *test-data*.
  "
  (set! *test-data* (read-data file))
  #t)

(define (read-data file)
  "
  Purpose:	Read list of expressions from a file.
  Remark:       Complains if expressions vary in length.
  Returns:	A list of expressions.
  "
  (let ((result '())
	(record-length 0)
	(count 0))
    (define (read-expression)
      (let ((e (read)))
	(if (eof-object? e)
	    'done
	    (begin
	      (if (zero? count)
		  (set! record-length (length e))
		  (if (not (= (length e) record-length))
		      (format #t "~%Record ~a is wrong length" count)))
	      (set! result (cons e result))
	      (set! count (1+ count))
	      (read-expression)))))
    (with-input-from-file file read-expression)
    result))

;;;; IDENTIFICATION

(define (identify unknown)
  "
  Purpose:	To perform a diagnosis using an identification tree.
  Argument:	A list of a diagnosis and various attribute values
  Returns:	The best guess for an unknown's diagnosis.
  "
 (let* ((result (find-answer unknown *tree* 0))
	(diagnosis (answer-diagnosis result))
	(supporter-count (answer-support result)))
      (format #t "~%The winner is ~a according to ~a example~a."
	      diagnosis supporter-count (if (= 1 supporter-count) "" "s"))
      diagnosis))

(define (find-answer unknown tree level)
  "
  Purpose:	Find answer using identification tree.
  Remarks:	This is less complicated than the kd tree identifier
                because there is no backup.
  "
  (let* ((dimension (branching-node-dimension tree))
	 (threshold (branching-node-threshold tree))
	 (projection (list-ref (record-attributes unknown) dimension))
	 (next-node #f))
    ;;Decide which branch has won:
    (if (> projection threshold)
	(begin
	  (say *verbose* "~%~aTurn toward large numbers at level ~a(~a): ~
		       ~a is greater than ~a."
	       (indent level)
	       level dimension projection
	       (branching-node-threshold tree))
	  (set! next-node (branching-node-right-branch tree)))
	(begin (say *verbose* "~%~aTurn toward small numbers at level ~a(~a): ~
		     ~a is less than or equal to ~a."
		    (indent level)
		    level dimension projection 
		    (branching-node-threshold tree))
	       (set! next-node (branching-node-left-branch tree))))
    ;;If the winning branch is an branching node ...
    (if (branching-node? next-node)
	;; Then find the closest neighbor by calling FIND-ANSWER recursively:
	(find-answer unknown next-node (1+ level))
	;; Otherwise the winning branch is a terminal node:
	(package-answer next-node))))

(define (package-answer node)		;Constructor
  (list (terminal-node-identity node)
	(length (terminal-node-samples node))))

(define answer-diagnosis first)		;Getters

(define answer-support second)


;;;; DEBUGGING PROCEDURES

(define (say switch #!rest args)
  "
  Purpose:	Display stuff if switch is set to #t.
  "
  (if switch (apply format #t args)))

(define *verbose* #f)
