;;;; -*- mode:Scheme -*- ;;;;

;;;; PRODUCE FASTER CODE

(declare (usual-integrations))

;;;; USE THE FORMAT LIBRARY

(load-option 'format)

;;;; IMITATATE COMMONLISP

(define (remove-duplicates l)
  "
  Purpose:	   Remove duplicates in a list.
  Sample argument: (a b c b d)
  Sample value:    (a b c d)
  "
  (cond ((null? l) ())
	((member (first l) (cdr l)) (remove-duplicates (cdr l)))
	(else (cons (first l) (remove-duplicates (cdr l))))))

;;;; ACCESS FUNCTIONS FOR SAMPLES

(define (record-diagnosis record)
  (let ((d (first record)))
    (cond ((eq? d 'primary) 1)
	  (else 0))))

(define (record-attributes record)
  (let ((attributes (cdr record)))
    (map (lambda (x) (* x 1.0))
	 (list
	  (list-ref attributes 2)
	  (list-ref attributes 7)
	  (list-ref attributes 15)
	  (list-ref attributes 19)))))

;;;; GLOBAL VARIABLES

(define *initializing-data*)  ; Assigned to a list of records used to establish
                              ; RBF "centers," and optionally, initial weights

(define *training-data*)      ; Assigned to a list of training records

(define *test-data*)          ; Assigned to a list of test records

(define *net*)                ; Assigned to an RBF net by MAKE-NET

(define *sigma*)              ; The width of the Gaussians used in the RBF net

(define *compute-weights*)    ; If #t, weights are computed such that
                              ; perfect results are obtained from the samples
                              ; bound to *initialization-data*; otherwise,
                              ; initial weights are set to 0

(define *rate*)               ; The multiplicative rate factor used in gradient
                              ; ascent

(define *steps*)              ; Assigned to the total number of training steps
                              ; for the net


;;;; DATA STRUCTURE

; Each RBF net is a structure with slots for weights, centers, and sigmas.
; In this implementation, the sigmas for all nodes are the same.
; You need not study this part of the code.

(define rbf-net
  (make-record-type "rbf-net" '(centers weights sigmas)))

(define make-rbf-net (record-constructor rbf-net))

(define rbf-net-weights
  (record-accessor rbf-net 'weights))

(define rbf-net-centers
  (record-accessor rbf-net 'centers))

(define rbf-net-sigmas
  (record-accessor rbf-net 'sigmas))

(define set-rbf-net-weights
  (record-modifier rbf-net 'weights))

(define set-rbf-net-centers
  (record-modifier rbf-net 'centers))


;;;; NET CONSTRUCTION

(define (make-net samples)
  "
  Purpose:	Create an rbf net.
  Arguments:	'(o1 x11 x12 x13)	;First sample, with desired output. 
		'(o2 x21 x22 x23)
	         ...
	        '(on xn1 xn2 xn3)	;Nth sample, with desired output. 
                                        ; os are outputs, xs are inputs.
  Returns:	Rbf net structure.
  "
  ;; The initial number of training steps is 0:
  (set! *steps* 0)
  ;; Fetch the net inputs and outputs from the samples;
  ;; All corresponding sigmas are the same:
  (let* ((inputs (map record-attributes samples))
	 (outputs (map record-diagnosis samples))
	 (sigmas (map (lambda (x) *sigma*) samples)))
    (set! *net*
	  ;; The real work is done by make-rbf-net; it has three arguments--
          ;; the inputs, the initial weights, and the sigmas:
	  (make-rbf-net
	   inputs
	   (if *compute-weights*
	       ;; Solve some simultaneous equations if initial weights
	       ;; are to be computed:
	       (let ((equations '()))
		 (do ((centers inputs (cdr centers))
		      (outputs outputs (cdr outputs)))
		     ((null? centers))
		   (let ((equation '()))
		     (for-each
		      (lambda (x)
			(set! equation
			      (cons
			       (hidden-output x (first centers) *sigma*)
			       equation)))
			       inputs)
		     (set! equation (cons (first outputs) equation))
		     (set! equations (cons (reverse equation) equations))))
		 (solve-equations (reverse equations)))
	       ;; Otherwise, all initial weights are 0:
	       (map (lambda (x) 0.0) inputs))
	   sigmas))))

;;; NET OUTPUT

(define (compute-net-output record)
  "
  Purpose:	Compute the output produced by a particular record.
  Arguments:	The record.
  Returns:	The output.
  "
  ;; Add up results for each node:
  (reduce +
	  0.0
	  (map (lambda (center weight sigma)
		 ;; For each node, compute output and multiply by weight:
		 (let ((w weight)
		       (h (hidden-output (record-attributes record)
					  center
					  sigma)))
		   ;; This hopeless crock of a test is needed because
		   ;; some hardware traps mysteriously if the numbers 
		   ;; are extremely small, but not zero: 
		   (if (and (> (abs w) 10e-10) (> (abs h) 10e-10))
		       (* w h)
		       0.0)))
	       (rbf-net-centers *net*)
	       (rbf-net-weights *net*)
	       (rbf-net-sigmas *net*))))

(define (hidden-output vector center sigma)
  "
  Purpose:	Compute hidden-node output.
  Arguments:	The input vector, the center, and sigma.
  Returns:	The node output.
  "
  (gauss-of-x2 (vector-norm-squared vector center) sigma))

(define (gauss-of-x2 x sigma)
  (let ((argument (/ x (* -2.0 sigma))))
    ;; See "crock" note above for explanation of the following:
    (cond ((< (abs argument) 1e-10) 1.0)
	  ((> (abs argument) 744.0) 0.0)
	  (else (exp argument)))))

;;;; TRAINING

(define (train-weights steps)
  "
  Purpose:	Train an already-created rbf net.
  Arguments:	The number of training steps to be taken.
  Returns:	#t
  "
  (do ((count 0 (1+ count)))
      ((= count steps) #t)
    ;; Increment step counter:
    (set! *steps* (1+ *steps*))
    ;; Get a random sample and compute weight changes using it:
    (let* ((sample (random-training-sample))
	   (derivatives (compute-weight-derivatives sample))
	   (deltas (vector-times-scaler *rate* derivatives)))
      ;; Adjust weights:
      (set-rbf-net-weights
       *net*
       (vector-sum deltas (rbf-net-weights *net*))))))

(define (train-weights-and-centers steps)
  "
  Purpose:	Train an already-created rbf net.
  Arguments:	The number of training steps to be taken.
  Returns:	#t
  Remark:       Not used in the homework, except optionally.
  "
  (do ((count 0 (1+ count)))
;;      ((= count steps) net)
      ((= count steps) #t)
    ;; Increment step counter:
    (set! *steps* (1+ *steps*))
    ;; Get a random sample and compute weight and center changes using it:
    (let* ((sample (random-training-sample)))
      ;; Adjust weights:
      (set-rbf-net-weights
       *net*
       (vector-sum
	(vector-times-scaler *rate*
			     (compute-weight-derivatives sample))
	(rbf-net-weights *net*)))
      ;; Adjust centers; this is more complicated because each center
      ;; is itself a vector:
      (set-rbf-net-centers
       *net*
       (map (lambda (v1 v2) (vector-sum v1 v2))
	    (map (lambda (v) (vector-times-scaler *rate* v))
		 (compute-center-derivatives sample))
	    (rbf-net-centers *net*))))))

;;;; COMPUTE ERRORS AND DERIVATIVES

(define (compute-weight-derivatives sample)
  "
  Remarks:	Embodies result of differentiation of performance
                function wrt weights.
  Note:         Inefficient beyond description; computations repeated.
  "
  (map (lambda (center sigma)
	 (* (- (record-diagnosis sample) (compute-net-output sample))
	    (hidden-output (record-attributes sample) center sigma)))
       (rbf-net-centers *net*)
       (rbf-net-sigmas *net*)))

(define (compute-center-derivatives sample)
  "
  Remarks:	Embodies result of differentiation of performance
                function wrt centers.
  Note:         Inefficient beyond description; computations repeated.
  "
  (map (lambda (weight center sigma)
	 (do ((n 0 (1+ n))
	      (result ()))
	     ((= n (length center)) (reverse result))
	   (set! result
		 (cons (* (/ weight sigma)
			  (- (list-ref (record-attributes sample) n)
			     (list-ref center n))
			  (- (record-diagnosis sample)
			     (compute-net-output sample))
;;;			  (hidden-output inputs center sigma))
			  (hidden-output 
			   (record-attributes sample) center sigma))
		       result))))
       (rbf-net-weights *net*)
       (rbf-net-centers *net*)
       (rbf-net-sigmas *net*)))

      
;;;; AUXILIARY VECTOR PROCEDURES

(define (vector-times-scaler s v) (map (lambda (e) (* s e)) v))

(define (vector-sum x y) (map + x y))

(define (vector-difference x y) (map - x y))

(define (vector-norm-squared x y)
  (reduce + 0 (map (lambda (e) (* e e))
		   (vector-difference x y))))

(define (vector-norm x y)
  (sqrt (vector-norm-squared x y)))

;;;; DATA READERS

(define (read-initializing-data file)
  "
  Purpose:	Read training data.  Assign *training-data*.
  "
  (set! *initializing-data* (read-data file))
  (format #t "~%There are ~a net-initializing records"
	  (length *initializing-data*))
  #t)

(define (read-training-data file)
  "
  Purpose:	Read training data.
  "
  (set! *training-data* (read-data file))
  (format #t "~%There are ~a training records" (length *training-data*))
  #t)

(define (random-training-sample)
   (list-ref *training-data* (random (length *training-data*))))

(define (read-test-data file)
  "
  Purpose:	Read test data.
  "
  (set! *test-data* (read-data file))
  (format #t "~%There are ~a test records" (length *test-data*))
  #t)

(define (read-data file)
  "
  Purpose:	Read list of expressions from a file.
  Remark:       Complains if expressions vary in length.
  Returns:	A list of expressions.
  "
  (let ((result '())
	(record-length 0)
	(field-count 0)
	(record-count 0))
    (define (read-expression)
      (let ((e (read)))
	(if (eof-object? e)
	    'done
	    (begin
	      (if (zero? field-count)
		  (set! record-length (length e))
		  (if (not (= (length e) record-length))
		      (format #t "~%Record ~a is wrong length" field-count)))
	      (set! result (cons e result))
	      (set! field-count (1+ field-count))
	      (set! record-count (1+ record-count))
	      (read-expression)))))
    (with-input-from-file file read-expression)
    result))

;;;; EVALUATE PERFORMANCE

(define (evaluate)
  (let ((pos-total 0) (pos-winners 0) (neg-total 0) (neg-winners 0))
    (format #t "~%Evaluating performance after ~a steps, please wait ..."
	    *steps*)
    (for-each
     (lambda (sample)
       (let ((desired-output (record-diagnosis sample))
	     (actual-output (compute-net-output sample)))
	 (cond ((eq? desired-output 1)
		(set! pos-total (1+ pos-total))
		(if (> actual-output 0.5)
		    (set! pos-winners (1+ pos-winners))
		    (format #t "~%Lost on ~a: ~a"
			    (first sample)
			    (record-attributes sample))))
	       (else
		(set! neg-total (1+ neg-total))
		(if (< actual-output 0.5)
		    (set! neg-winners (1+ neg-winners)))))))
     *test-data*)
    (format #t "~%~a out of ~a (~a%) positives recognized"
	    pos-winners pos-total (floor (* 100.0 (/ pos-winners pos-total))))
    (format #t "~%~a out of ~a (~a%) non-positives recognized"
	    neg-winners neg-total (floor (* 100.0 (/ neg-winners neg-total))))
    ))






